/*
    This file is a part of libcds - Concurrent Data Structures library

    (C) Copyright Maxim Khizhinsky (libcds.dev@gmail.com) 2006-2017

    Source code repo: http://github.com/khizmax/libcds/
    Download: http://sourceforge.net/projects/libcds/files/

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, this
      list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
    IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
    DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
    SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
    OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "map_type.h"

namespace map {


    class Map_InsDelFind: public cds_test::stress_fixture
    {
    public:
        static size_t s_nMapSize;           // initial map size

        static size_t s_nPassCount;
        static size_t s_nBronsonAVLTreeMapPassCount;

        static size_t s_nHpEllenBinTreeMapPassCount;
        static size_t s_nHpFeldmanPassCount;
        static size_t s_nHpMichaelMapPassCount;
        static size_t s_nHpSkipListMapPassCount;
        static size_t s_nHpSplitListMapPassCount;

        static size_t s_nRcuEllenBinTreeMapPassCount;
        static size_t s_nRcuFeldmanPassCount;
        static size_t s_nRcuMichaelMapPassCount;
        static size_t s_nRcuSkipListMapPassCount;
        static size_t s_nRcuSplitListMapPassCount;


        static size_t s_nThreadCount;       // thread count
        static size_t s_nMaxLoadFactor;     // maximum load factor
        static unsigned int s_nInsertPercentage;
        static unsigned int s_nDeletePercentage;
        static unsigned int s_nDuration;    // test duration, seconds

        static size_t s_nCuckooInitialSize;         // initial size for CuckooMap
        static size_t s_nCuckooProbesetSize;        // CuckooMap probeset size (only for list-based probeset)
        static size_t s_nCuckooProbesetThreshold;   // CuckooMap probeset threshold (o - use default)

        static size_t s_nFeldmanMap_HeadBits;
        static size_t s_nFeldmanMap_ArrayBits;

        static size_t  s_nLoadFactor;  // current load factor

        static void SetUpTestCase();
        //static void TearDownTestCase();

    public:
        enum actions
        {
            do_find,
            do_insert,
            do_delete
        };
        static const unsigned int c_nShuffleSize = 100;
        static actions s_arrShuffle[c_nShuffleSize];

    protected:
        typedef size_t  key_type;
        typedef size_t  value_type;

        template <class Map>
        class Worker: public cds_test::thread
        {
            typedef cds_test::thread base_class;
            Map&     m_Map;

        public:
            size_t  m_nInsertSuccess = 0;
            size_t  m_nInsertFailed = 0;
            size_t  m_nDeleteSuccess = 0;
            size_t  m_nDeleteFailed = 0;
            size_t  m_nFindSuccess = 0;
            size_t  m_nFindFailed = 0;

        public:
            Worker( cds_test::thread_pool& pool, Map& map )
                : base_class( pool )
                , m_Map( map )
            {}

            Worker( Worker& src )
                : base_class( src )
                , m_Map( src.m_Map )
            {}

            virtual thread * clone()
            {
                return new Worker( *this );
            }

            typedef std::pair< key_type const, value_type > map_value_type;

            struct update_functor {
                template <typename Q>
                void operator()( bool /*bNew*/, map_value_type& /*cur*/, Q const& /*val*/ ) const
                {}

                // FeldmanHashMap
                void operator()( map_value_type& /*cur*/, map_value_type* /*old*/) const
                {}

                // MichaelMap
                void operator()( bool /*bNew*/, map_value_type& /*cur*/ ) const
                {}

                // BronsonAVLTreeMap
                void operator()( bool /*bNew*/, key_type /*key*/, value_type& /*val*/ ) const
                {}
            };

            virtual void test()
            {
                Map& rMap = m_Map;
                size_t pass_count = Map_InsDelFind::s_nPassCount;

                size_t nInsertedNum = 0;
                size_t nUninsertedNum = 0;
                for (size_t count = 0; count < pass_count; count++) {
                  bool shouldUpdate = true;
                  for (size_t i = 0; i < s_nMapSize; ++i) {
                    // The number to operate on the map.
                    size_t n = i;
                    // Insert
                    if (i % s_nInsertPercentage == 1) {
                      nInsertedNum++;
                      if (!shouldUpdate) {
                        if (rMap.insert(n, n))
                          ++m_nInsertSuccess;
                        else
                          ++m_nInsertFailed;
                        shouldUpdate = true;
                      } else {
                        if (rMap.update(n, update_functor(), true).first)
                          ++m_nInsertSuccess;
                        else
                          ++m_nInsertFailed;
                        shouldUpdate = false;
                      }
                    } else {
                      nUninsertedNum++;
                    }
                    // Find
                    if (rMap.contains(n))
                      ++m_nFindSuccess;
                    else
                      ++m_nFindFailed;
                    // Delete
                    if (i % s_nInsertPercentage == 1) {

                      if (rMap.erase(n))
                        ++m_nDeleteSuccess;
                      else
                        ++m_nDeleteFailed;
                      break;
                    }
                  }
                }
                EXPECT_EQ(nInsertedNum, m_nFindSuccess);
                EXPECT_EQ(nUninsertedNum, m_nFindFailed);
                EXPECT_EQ(m_nDeleteSuccess, nInsertedNum);
                EXPECT_EQ(m_nDeleteFailed, 0);
            }
        };

    protected:
        template <class Map>
        void do_test( Map& testMap )
        {
            typedef Worker<Map> worker;
            cds_test::thread_pool& pool = get_pool();
            std::unique_ptr<worker> worker_thrd(new worker(pool, testMap));
            worker_thrd->test();

            testMap.clear();
            EXPECT_TRUE( testMap.empty());
            additional_cleanup( testMap );
        }

        template <class Map>
        void run_test()
        {
          std::unique_ptr<Map> testMap(new Map(*this));
          do_test(*testMap);
        }

        template <class Map>
        void run_bronson_avl_tree() {
          Map_InsDelFind::s_nPassCount =
              Map_InsDelFind::s_nBronsonAVLTreeMapPassCount;
          run_test<Map>();
        }

        template <class Map>
        void run_ellen_bin_tree_hp() {
          Map_InsDelFind::s_nPassCount =
              Map_InsDelFind::s_nHpEllenBinTreeMapPassCount;
          run_test<Map>();
        }

        template <class Map>
        void run_skip_list_hp() {
          Map_InsDelFind::s_nPassCount =
              Map_InsDelFind::s_nHpSkipListMapPassCount;
          run_test<Map>();
        }

        template <class Map>
        void run_feldman_hp() {
          Map_InsDelFind::s_nPassCount =
              Map_InsDelFind::s_nHpFeldmanPassCount;
          run_test<Map>();
        }

        template <class Map>
        void run_ellen_bin_tree_rcu() {
          Map_InsDelFind::s_nPassCount =
              Map_InsDelFind::s_nRcuEllenBinTreeMapPassCount;
          run_test<Map>();
        }

        template <class Map>
        void run_skip_list_rcu() {
          Map_InsDelFind::s_nPassCount =
              Map_InsDelFind::s_nRcuSkipListMapPassCount;
          run_test<Map>();
        }

        template <class Map>
        void run_feldman_rcu() {
          Map_InsDelFind::s_nPassCount =
              Map_InsDelFind::s_nRcuFeldmanPassCount;
          run_test<Map>();
        }
    };

    class Map_InsDelFind_LF: public Map_InsDelFind
        , public ::testing::WithParamInterface<size_t>
    {
    public:
        template <class Map>
        void run_test()
        {
            s_nLoadFactor = GetParam();
            propout() << std::make_pair( "load_factor", s_nLoadFactor );
            Map_InsDelFind::run_test<Map>();
        }

				template <class Map>
        void run_michael_hp() {
          Map_InsDelFind::s_nPassCount =
              Map_InsDelFind::s_nHpMichaelMapPassCount;
          Map_InsDelFind_LF::run_test<Map>();
        }

        template <class Map>
        void run_split_list_hp() {
          Map_InsDelFind::s_nPassCount =
              Map_InsDelFind::s_nHpSplitListMapPassCount;
          Map_InsDelFind_LF::run_test<Map>();
        }

        template <class Map>
        void run_michael_rcu() {
          Map_InsDelFind::s_nPassCount =
              Map_InsDelFind::s_nRcuMichaelMapPassCount;
          Map_InsDelFind_LF::run_test<Map>();
        }

        template <class Map>
        void run_split_list_rcu() {
          Map_InsDelFind::s_nPassCount =
              Map_InsDelFind::s_nRcuSplitListMapPassCount;
          Map_InsDelFind_LF::run_test<Map>();
        }

        static std::vector<size_t> get_load_factors();
    };

} // namespace map
